//////////////////////////////////////////////////////////////////////
// concat.c - A simple picture manipulator to place several bitmaps on a grid
// If you do something with this code, all I ask is that you tell me about it
// I accept no responsibility for the consequences of using this code
// Copyright Joe Kopena 1997-98
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>

// This is all very shabbily written, but... I was feeling very lazy and didn't want to
// spend a lot of time on this.

BITMAP **bitmaps;
BITMAP *result;
PALETTE pal;
char filename[80];
unsigned char gColor, gSize;
int width, height, sX, sY;
char fWidth, fHeight, noBorder;
char filesList;
int numBitmaps;

void showUsage(void) {

   printf("Compose usage:\n");
   printf("     compose filename [-gColor n][-gSize n][-w n][-h n][-sx n][-sy n]\n");
   printf("                      [-noBorder]\n");
   printf("          filename is a valid filename, with the extension .pcx, .bmp or .tga\n");
   printf("          -gColor allows you to specify the grid (background) color\n");
   printf("          -gSize allows you to specify the grid size\n");
   printf("          -w forces a width for the result file\n");
   printf("          -h forces a height for the result file\n");
   printf("          -sx forces the starting x co-ord\n");
   printf("          -sy forces the starting y co-ord\n");
   printf("          -noBorder removes the right and bottom grid lines");

// end showUsage
}

void getCmdLineStuff(int argc, char *argv[]) {
int i, t;

   if (argc < 2) {
      showUsage();
      exit(-1);
   }

   fWidth = 0;
   fHeight = 0;
   width = 0;
   height = 0;
   sX = 0;
   sY = 0;
   gColor = 255;
   gSize = 1;
   noBorder = 0;

   strcpy(filename, argv[1]); // First parameter is the resulting filename
   if (!strstr(filename, ".bmp") && !strstr(filename, ".pcx") && !strstr(filename, ".tga")) {
      showUsage();
      exit(-1);
   }

   t = 2;   // t tracks where the filenames begin
   for (i = 2; i < argc; i++) {
       if (!stricmp(argv[i], "-gColor") && i < argc-1) {   // Set grid color
          gColor = atoi(argv[i+1]);
          t += 2;
       } else if (!stricmp(argv[i], "-w") && i < argc-1) { // Set result width
         width = atoi(argv[i+1]);
         fWidth = 1;
          t += 2;
       } else if (!stricmp(argv[i], "-h") && i < argc-1) { // Set result height
         height = atoi(argv[i+1]);
         fHeight = 1;
          t += 2;
       } else if (!stricmp(argv[i], "-gSize") && i < argc-1) { // Set grid size
         gSize = atoi(argv[i+1]);
         t += 2;
       } else if (!stricmp(argv[i], "-sx") && i < argc - 1) { // Starting pos
         sX = atoi(argv[i+1]);
         t += 2;
       } else if (!stricmp(argv[i], "-sy") && i < argc - 1) { // Starting pos
         sY = atoi(argv[i+1]);
         t += 2;
       } else if (!stricmp(argv[i], "-noBorder")) {
         noBorder = 1;
       }

   } // for i

   filesList = t; // lazy :)

   width += sX;
   height += sY;

// end getCmdLineStuff
}

void getBitmaps(int argc, char *argv[]) {
int i, x, y, maxHeight;

   numBitmaps = argc - filesList;
   if (!numBitmaps) {
      printf("Warning, no files specified\n");
      return;
   }

   bitmaps = malloc(sizeof(BITMAP *) * numBitmaps);
   if (!bitmaps) {
      printf("Not enough memory for bitmaps holder\n");
      exit(-1);
   }

   printf("Adding %i files:\n", numBitmaps);
   for (i=0; i < numBitmaps; i++) {

       bitmaps[i] = load_bitmap(argv[filesList+i], pal);
       if (!bitmaps[i]) {
          printf("Not enough for bitmap\n");
          exit(-1);
       }

//       printf("\t #%i: %s (%i, %i)\n", i+1, argv[filesList+i], bitmaps[i]->w, bitmaps[i]->h);

   } // for i

//   if (!fHeight && noBorder) {
//      height += maxHeight;
//      height -= gSize;
//   }

//   if (!fWidth && noBorder)
//      width -= gSize;      // Remove right border
   x = sX;
   y = sY;
   maxHeight = 0;

   for (i = 0; i < numBitmaps; i++) {
       if (bitmaps[i]->h > maxHeight)
          maxHeight = bitmaps[i]->h;

       if (x + bitmaps[i]->w + ((noBorder)?0:gSize) > width) {
          if (!fWidth)
             width = x + bitmaps[i]->w + ((noBorder)?0:gSize);
          else {
               if (noBorder) {
                  x = sX;
                  y += maxHeight + gSize;
                  maxHeight = 0;
               } else {
                 if (x + bitmaps[i]->w > width) {
                     x = sX;
                     y += maxHeight + gSize;
                     maxHeight = 0;
                  } else printf("Warning: No rightmost border space\n");
               } // !noBorder
          } // fWidth
       } // width

       if (height < y + bitmaps[i]->h + ((noBorder)?0:gSize)) {
          if (fHeight)
             printf("Warning: Bitmap does not fit\n");
          else {
               height = y + bitmaps[i]->h + ((noBorder)?0:gSize);
          }
       } // height

//       blit(bitmaps[i], result, 0, 0, x, y, bitmaps[i]->w, bitmaps[i]->h);
//       printf("(x, y):(%i, %i); (w, h):(%i, %i)\n", x, y, width, height);
       printf("\t #%i: %s (%i, %i) at (%i, %i)\n", i+1, argv[filesList+i], bitmaps[i]->w, bitmaps[i]->h, x, y);

       x += bitmaps[i]->w + gSize;

   } // for i

   if (noBorder && !fHeight)
      height -= gSize;

// end getBitmaps
}

void writeBitmaps(void) {
int i, x, y, mHeight = 1;

   x = sX;
   y = sY;

   for (i = 0; i < numBitmaps; i++) {
       if (bitmaps[i]->h > mHeight)
          mHeight = bitmaps[i]->h;

       if (x + bitmaps[i]->w > width) {
          x = sX;
          y += mHeight + gSize;
          mHeight = 0;
       }

       blit(bitmaps[i], result, 0, 0, x, y, bitmaps[i]->w, bitmaps[i]->h);
//       printf("(x, y):(%i, %i)\n", x, y);

       x += bitmaps[i]->w + gSize;

   } // for i

// end writeBitmaps
}

int main(int argc, char *argv[]) {
int err = 0, i;

   memcpy(pal, desktop_palette, sizeof(PALETTE));

   getCmdLineStuff(argc, argv);

   getBitmaps(argc, argv);

   result = create_bitmap(width, height);
   if (!result) {
      printf("Not enough memory for resulting bitmap\n");
      exit(-1);
   }
   clear_to_color(result, gColor);

   writeBitmaps();

   save_bitmap(filename, result, pal);
   destroy_bitmap(result);

   for (i = 0; i < numBitmaps; i++)
       destroy_bitmap(bitmaps[i]);
   free(bitmaps);

// end main
return err;
}

